import re
import hashlib
import logging
import os

import requests
import urllib3
from bs4 import BeautifulSoup
from retrying import retry

import enum

urllib3.disable_warnings()

# URL-DIRECT - openly accessible paper
# URL-NON-DIRECT - pay-walled paper
# PMID - PubMed ID
# DOI - digital object identifier
IDClass = enum.Enum("identifier", ["URL-DIRECT", "URL-NON-DIRECT", "PMD", "DOI"])

DEFAULT_USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.3 Safari/605.1.15"


class IdentifierNotFoundError(Exception):
    "Error for when the identifier wasn't found at any Sci-Hub url"

    pass


class SiteAccessError(Exception):
    pass


class CaptchaNeededError(SiteAccessError):
    pass


class SciHub(object):
    """
    Sci-Hub class can search for papers on Google Scholar
    and fetch/download papers from sci-hub.io
    """

    def __init__(self, user_agent=DEFAULT_USER_AGENT):
        self.sess = requests.Session()
        self.sess.headers = {
            "User-Agent": user_agent,
        }
        self.available_base_url_list = self._get_available_scihub_urls()

        self.base_url = self.available_base_url_list[0] + "/"

    def _get_available_scihub_urls(self):
        """
        Finds available Sci-Hub urls via https://sci-hub.now.sh/
        """

        # NOTE: This misses some valid URLs. Alternatively, we could parse
        # the HTML more finely by navigating the parsed DOM, instead of relying
        # on filtering. That might be more brittle in case the HTML changes.
        # Generally, we don't need to get all URLs.
        scihub_domain = re.compile(r"^http[s]*://sci.hub", flags=re.IGNORECASE)
        urls = []
        res = requests.get("https://sci-hub.now.sh/")
        s = self._get_soup(res.content)
        text_matches = s.find_all("a", href=True, string=re.compile(scihub_domain))
        href_matches = s.find_all("a", re.compile(scihub_domain), href=True)
        full_match_set = set(text_matches) | set(href_matches)
        for a in full_match_set:
            if "sci" in a or "sci" in a["href"]:
                urls.append(a["href"])
        return urls

    def set_proxy(self, proxy):
        """
        set proxy for session
        :param proxy_dict:
        :return:
        """
        if proxy:
            self.sess.proxies = {
                "http": proxy,
                "https": proxy,
            }

    def _change_base_url(self):
        if len(self.available_base_url_list) <= 1:
            logging.error("Ran out of valid Sci-Hub urls")
            raise IdentifierNotFoundError()
        del self.available_base_url_list[0]
        self.base_url = self.available_base_url_list[0] + "/"

        logging.info("Changing URL to {}".format(self.available_base_url_list[0]))

    def download(self, identifier, destination="", path=None) -> dict[str, str] | None:
        """
        Downloads a paper from Sci-Hub given an indentifier (DOI, PMID, URL).
        Currently, this can potentially be blocked by a captcha if a certain
        limit has been reached.
        """
        try:
            data = self.fetch(identifier)

            # TODO: allow for passing in name
            if data:
                self._save(
                    data["pdf"],
                    os.path.join(destination, path if path else data["name"]),
                )
            return data
        except IdentifierNotFoundError:
            logging.error(f"Failed to find identifier {identifier}")

    @retry(
        wait_random_min=100,
        wait_random_max=1000,
        stop_max_attempt_number=20,
        retry_on_exception=lambda e: not (
            isinstance(e, IdentifierNotFoundError) or isinstance(e, IndexError)
        ),
    )
    def fetch(self, identifier) -> dict | None:
        """
        Fetches the paper by first retrieving the direct link to the pdf.
        If the indentifier is a DOI, PMID, or URL pay-wall, then use Sci-Hub
        to access and download paper. Otherwise, just download paper directly.
        """
        logging.info(f"Looking for {identifier}")
        try:
            # find the url to the pdf for a given identifier
            url = self._get_direct_url(identifier)
            logging.info(f"Found potential source at {url}")

            # verify=False is dangerous but Sci-Hub.io
            # requires intermediate certificates to verify
            # and requests doesn't know how to download them.
            # as a hacky fix, you can add them to your store
            # and verifying would work. will fix this later.
            # NOTE(ben): see this SO answer: https://stackoverflow.com/questions/27068163/python-requests-not-handling-missing-intermediate-certificate-only-from-one-mach
            res = self.sess.get(url, verify=True)

            if res.headers["Content-Type"] != "application/pdf":
                logging.error(
                    f"Couldn't find PDF with identifier {identifier} at URL {url}, changing base url..."
                )
                raise SiteAccessError("Couldn't find PDF")
            else:
                return {
                    "pdf": res.content,
                    "url": url,
                    "name": self._generate_name(res),
                }

        except Exception as e:
            if len(self.available_base_url_list) < 1:
                raise IdentifierNotFoundError("Ran out of valid Sci-Hub urls")
            logging.info(
                f"Cannot access source from {self.available_base_url_list[0]}: {e}, changing base URL..."
            )
            self._change_base_url()
            raise SiteAccessError from e

    def _get_direct_url(self, identifier: str) -> str:
        """
        Finds the direct source url for a given identifier.
        """
        id_type = self._classify(identifier)

        if id_type == IDClass["URL-DIRECT"]:
            return identifier
        else:
            return self._search_direct_url(identifier)

    def _search_direct_url(self, identifier) -> str:
        """
        Sci-Hub embeds papers in an iframe. This function finds the actual
        source url which looks something like https://moscow.sci-hub.io/.../....pdf.
        """

        while True:
            res = self.sess.get(self.base_url + identifier, verify=True)
            s = self._get_soup(res.content)
            iframe = s.find("iframe")

            if iframe:
                src = iframe.get("src")
                if isinstance(src, list):
                    src = src[0]
                if src.startswith("//"):
                    return "http:" + src
                else:
                    return src

            else:
                self._change_base_url()

    def _classify(self, identifier) -> IDClass:
        """
        Classify the type of identifier:
        url-direct - openly accessible paper
        url-non-direct - pay-walled paper
        pmid - PubMed ID
        doi - digital object identifier
        """
        if identifier.startswith("http") or identifier.startswith("https"):
            if identifier.endswith("pdf"):
                return IDClass["URL-DIRECT"]
            else:
                return IDClass["URL-NON-DIRECT"]
        elif identifier.isdigit():
            return IDClass["PMID"]
        else:
            return IDClass["DOI"]

    def _save(self, data, path):
        """
        Save a file give data and a path.
        """
        try:
            logging.info(f"Saving file to {path}")

            with open(path, "wb") as f:
                f.write(data)
        except Exception as e:
            logging.error(f"Failed to write to {path} {e}")
            raise e

    def _get_soup(self, html):
        """
        Return html soup.
        """
        return BeautifulSoup(html, "html.parser")

    def _generate_name(self, res):
        """
        Generate unique filename for paper by calcuating md5 hash of file
        contents.
        """
        pdf_hash = hashlib.md5(res.content).hexdigest()
        return f"{pdf_hash}" + ".pdf"



import argparse
import os
import sys
import logging


from scihub import SciHub
from parse import parse_file, format_output, parse_ids_from_text, id_patterns
import pdf2doi
import json

supported_fetch_identifier_types = ["doi", "pmid", "url", "isbn"]


def save_scihub(
    identifier: str, out: str, user_agent: str, name: str | None = None
) -> str:
    """
    find a paper with the given identifier and download it to the output
    directory. If given, name will be the name of the output file. otherwise
    we attempt to find a title from the PDF contents.
    """

    sh = SciHub(user_agent)
    logging.info(f"Attempting to download paper with identifier {identifier}")

    result = sh.download(identifier, out)
    if not result:
        return ""

    logging.info(f"Successfully downloaded paper with identifier {identifier}")

    logging.info("Finding paper title")
    pdf2doi.config.set("verbose", False)
    result_path = os.path.join(out, result["name"])

    try:
        result_info = pdf2doi.pdf2doi(result_path)
        validation_info = json.loads(result_info["validation_info"])

        title = validation_info.get("title")

        file_name = name if name else title
        if file_name:
            file_name += ".pdf"
            new_path = os.path.join(out, file_name)
            os.rename(result_path, new_path)
            logging.info(f"File renamed to {new_path}")
            return new_path
    except Exception as e:
        logging.error(f"Couldn't get paper title from PDF at {result_path}: {e}")

    return result_path


def parse(args) -> str:
    # if a path isn't passed or is empty, read from stdin
    if not (hasattr(args, "path") and args.path):
        return format_output(parse_ids_from_text(sys.stdin.read(), args.match))

    return format_output(parse_file(args.path, args.match), args.format)


def fetch(args) -> str:
    path = save_scihub(args.query, args.output, args.user_agent)
    if path:
        return path
    else:
        return "No paper found"


def main():
    name = "papers-dl"
    parser = argparse.ArgumentParser(
        prog=name,
        description="Download scientific papers from the command line",
    )

    from version import __version__

    parser.add_argument(
        "--version", "-V", action="version", version=f"{name} {__version__}"
    )

    parser.add_argument(
        "--verbose", "-v", action="store_true", help="increase verbosity"
    )

    subparsers = parser.add_subparsers()

    # FETCH
    parser_fetch = subparsers.add_parser(
        "fetch", help="try to download a paper with the given identifier"
    )

    parser_fetch.add_argument(
        "query",
        metavar="(DOI|PMID|URL)",
        type=str,
        help="the identifier to try to download",
    )

    parser_fetch.add_argument(
        "-o",
        "--output",
        metavar="path",
        help="optional output directory for downloaded papers",
        default=".",
        type=str,
    )

    parser_fetch.add_argument(
        "-A",
        "--user-agent",
        help="",
        default=None,
        type=str,
    )

    # PARSE
    parser_parse = subparsers.add_parser(
        "parse", help="parse identifiers from a file or stdin"
    )
    parser_parse.add_argument(
        "-m",
        "--match",
        metavar="type",
        help="the type of identifier to search for",
        type=str,
        choices=id_patterns.keys(),
        action="append",
    )
    parser_parse.add_argument(
        "-p",
        "--path",
        help="the path of the file to parse",
        type=str,
    )
    parser_parse.add_argument(
        "-f",
        "--format",
        help="the output format for printing",
        metavar="fmt",
        default="raw",
        choices=["raw", "jsonl", "csv"],
        nargs="?",
    )

    parser_fetch.set_defaults(func=fetch)
    parser_parse.set_defaults(func=parse)

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.ERROR)

    print(args.func(args))


if __name__ == "__main__":
    main()



from src.scihub import SciHub
import unittest


class TestSciHub(unittest.TestCase):
    def setUp(self):
        self.scihub = SciHub()

    def test_scihub_up(self):
        """
        Test to verify that `scihub.now.sh` is available
        """
        urls = self.scihub.available_base_url_list
        self.assertIsNotNone(urls, "Failed to find Sci-Hub domains")
        print(f"number of candidate urls: {len(urls)}")

    # NOTE: This test is flaky. Retrieval doesn't work consistently
    # def test_fetch(self):
    #     with open("tests/identifiers/ids.txt") as f:
    #         ids = f.read().splitlines()
    #         for id in ids:
    #             res = self.scihub.fetch(id)
    #             self.assertIsNotNone(res, f"Failed to fetch url from id {id}")
